#include <iostream>

#include "info.h"
#include "tree.h"
#include "calibd.h"
#include "calirotp.h"
#include "funs.h"

using std::cout;
using std::endl;


bool calirotp(tree::tree_p tnew, tree& x, xinfo& xi, dinfo& di, pinfo& pi, RNG& gen)
{
	tree::tree_p rotp,temp;
	tree::tree_cp xp;
	tree::npv subtold, subtnew, nbold, nbnew;
	double Qold_to_new, Qnew_to_old;
	unsigned int rdx=0;
	bool twowaystoinvert=false;
	double prinew=1.0,priold=1.0;
	size_t rotid;
	bool hardreject=false;
	std::vector<size_t> goodvars; //variables an internal node can split on

	pi.tproposal++;

	// Get rot nodes
	tree::npv rnodes;
	tnew->getrotnodes(rnodes);
	if(rnodes.size()==0) return false;  //no rot nodes so that's a reject.

	rdx = (unsigned int)floor(gen.uniform()*rnodes.size()); //which rotatable node will we rotate at?
	rotp = rnodes[rdx];
	rotid=rotp->nid();
	xp=x.getptr(rotid);

	int nwaysm1=0,nwaysm2=0,nwayss1=0,nwayss2=0;
	double pm1=1.0,pm2=1.0,ps1=1.0,ps2=1.0;
	if(rotp->isleft()) {
		if(rotp->v==rotp->p->v) //special case, faster to handle it direclty
		{
			rotright(rotp);
			rotp=tnew->getptr(rotid);
			delete rotp->r;
			temp=rotp->l;
			rotp->p->l=temp;
			temp->p=rotp->p;
			rotp->r=0;
			rotp->l=0;
			rotp->p=0;
			delete rotp;
			rotp=tnew->getptr(rotid);
			//pm1=pm2=ps1=ps2=1.0 in this case
		}
		else
		{
			rotright(rotp);
			rotp=tnew->getptr(rotid); //just in case the above changed the pointer.
			reduceleft(rotp,rotp->p->v,rotp->p->c);
			rotp=tnew->getptr(rotid); //just in case the above changed the pointer.
			reduceright(rotp->p->r,rotp->p->v,rotp->p->c);
			rotp=tnew->getptr(rotid); //just in case the above changed the pointer.
			splitleft(rotp->r,rotp->p->v,rotp->p->c);
			splitright(rotp->p->r->r,rotp->p->v,rotp->p->c);

			merge3(rotp->r,rotp->p->r->r,rotp->p->v,rotp->p->c,&nwayss1);
			ps1=1.0/nwayss1;

			merge3(rotp->l,rotp->p->r->l,rotp->p->v,rotp->p->c,&nwayss2);
			ps2=1.0/nwayss2;

			tree::tree_p tmerged=new tree;
			tmerged->p=rotp->p;

			merge3(rotp->p->r->l,rotp->p->r->r,rotp->p->r->v,rotp->p->r->c,&nwaysm1);
			pm1=1.0/nwaysm1;
			merge2(rotp->p->r->l,rotp->p->r->r,tmerged,rotp->p->r->v,rotp->p->r->c,gen);
			rotp->p->r->p=0;
			delete rotp->p->r;
			rotp->p->r=tmerged;

			tmerged=new tree;
			tmerged->p=rotp->p;

			merge3(rotp->l,rotp->r,rotp->v,rotp->c,&nwaysm2);
			pm2=1.0/nwaysm2;
			size_t v,c;
			v=rotp->v;
			c=rotp->c;
			merge2(rotp->l,rotp->r,tmerged,rotp->v,rotp->c,gen);
			rotp->p->l=tmerged;
			rotp->p=0;
			delete rotp;
			rotp=tnew->getptr(rotid);

		//end of merge code if rotp isleft.
			if( (rotp->v!=v && rotp->c!=c) && (rotp->p->r->v != v && rotp->p->r->c != c))
				hardreject=true;
			if(rotp->p->r->v==rotp->v && rotp->p->r->c==rotp->c && !isleaf(rotp->p->r))
				twowaystoinvert=true;

		}
	}
	else { //isright
		if(rotp->v==rotp->p->v) //special case, faster to handle it directly
		{
			rotleft(rotp);
			rotp=tnew->getptr(rotid);
			delete rotp->l;
			temp=rotp->r;
			rotp->p->r=temp;
			temp->p=rotp->p;
			rotp->r=0;
			rotp->l=0;
			rotp->p=0;
			delete rotp;
			rotp=tnew->getptr(rotid);
			//pm1=pm2=ps1=ps2=1.0 in this case
		}
		else
		{
			rotleft(rotp);
			rotp=tnew->getptr(rotid); //just in case the above changed the pointer.
			reduceleft(rotp->p->l,rotp->p->v,rotp->p->c);
			rotp=tnew->getptr(rotid); //just in case the above changed the pointer.
			reduceright(rotp,rotp->p->v,rotp->p->c);
			rotp=tnew->getptr(rotid); //just in case the above changed the pointer.
			splitleft(rotp->p->l->l,rotp->p->v,rotp->p->c);
			splitright(rotp->l,rotp->p->v,rotp->p->c);

			merge3(rotp->p->l->l,rotp->l,rotp->p->v,rotp->p->c,&nwayss1);
			ps1=1.0/nwayss1;

			merge3(rotp->p->l->r,rotp->r,rotp->p->v,rotp->p->c,&nwayss2);
			ps2=1.0/nwayss2;

			tree::tree_p tmerged=new tree;
			tmerged->p=rotp->p;

			merge3(rotp->p->l->l,rotp->p->l->r,rotp->p->l->v,rotp->p->l->c,&nwaysm1);
			pm1=1.0/nwaysm1;
			merge2(rotp->p->l->l,rotp->p->l->r,tmerged,rotp->p->l->v,rotp->p->l->c,gen);
			rotp->p->l->p=0;
			delete rotp->p->l;
			rotp->p->l=tmerged;

			tmerged=new tree;
			tmerged->p=rotp->p;

			merge3(rotp->l,rotp->r,rotp->v,rotp->c,&nwaysm2);
			pm2=1.0/nwaysm2;
			size_t v,c;
			v=rotp->v;
			c=rotp->c;
			merge2(rotp->l,rotp->r,tmerged,rotp->v,rotp->c,gen);
			rotp->p->r=tmerged;
			rotp->p=0;
			delete rotp;
			rotp=tnew->getptr(rotid);

		//end of merge code if rotp isright
			if( (rotp->v!=v && rotp->c!=c) && (rotp->p->l->v != v && rotp->p->l->c != c))
				hardreject=true;
			if(rotp->p->l->v==rotp->v && rotp->p->l->c==rotp->c && !isleaf(rotp->p->l))
				twowaystoinvert=true;
		}
	}

	// Calculate prior probabilities, we just need to use the subtree where the rotation occured of tnew and x.
	subtold.clear();
	subtnew.clear();
	xp->p->getnodes(subtold);
	rotp->p->getnodes(subtnew);

	for(size_t i=0;i<subtold.size();i++) {
		if(subtold[i]->l) { //interior node
			priold*=pi.alpha/pow(1.0 + subtold[i]->depth(),pi.beta);
			goodvars.clear();
			getinternalvars(subtold[i],xi,goodvars);
			priold*=1.0/((double)goodvars.size()); //prob split on v 
			priold*=1.0/((double)getnumcuts(subtold[i],xi,subtold[i]->v)); //prob split on v at c is 1/numcutpoints
		}
		else //terminal node
			priold*=(1.0-pi.alpha/pow(1.0 + subtold[i]->depth(),pi.beta));	
	}
	for(size_t i=0;i<subtnew.size();i++) {
		if(subtnew[i]->l) { //interior node
			prinew*=pi.alpha/pow(1.0 + subtnew[i]->depth(),pi.beta);
			goodvars.clear();
			getinternalvars(subtnew[i],xi,goodvars);
			prinew*=1.0/((double)goodvars.size()); //prob split on v
			prinew*=1.0/((double)getnumcuts(subtnew[i],xi,subtnew[i]->v)); //prob split on v at c is 1/numcutpoints
			if(getnumcuts(subtnew[i],xi,subtnew[i]->v)<1)
			{
				x.pr(true);
				tnew->pr(true);
			}
		}
		else //terminal node
			prinew*=(1.0-pi.alpha/pow(1.0 + subtnew[i]->depth(),pi.beta));	
	}

	Qold_to_new=1.0/((double)rnodes.size()); //proposal probability of rotating from x to tnew
	
	rnodes.clear();
	tnew->getrotnodes(rnodes);  //this is very inefficient, could make it much nicer later on.

	if(!twowaystoinvert)
		Qnew_to_old=1.0/((double)rnodes.size()); //proposal probability of rotating from tnew back to x
	else
		Qnew_to_old=2.0/((double)rnodes.size());

	// Calculate log integrated likelihoods for the subtree where the rotation occured of tnew and x.
	double lilold=0.0,lilnew=0.0;
	std::vector<sinfo> sold,snew;
	nbold.clear();
	nbnew.clear();
	x.getbots(nbold);
	tnew->getbots(nbnew);

	caliallsuff(x,xi,di,nbold,sold);
	caliallsuff(*tnew,xi,di,nbnew,snew);
	for(size_t i=0;i<nbold.size();i++) {
			if(nbold[i]->isrightchildofvsplit(pi.imdim))
				lilold += clil(0,sold[i].nf,0.0,sold[i].sf,pi.sigma,pi.sigmaf,pi.taud*pi.taud,pi.mud);
			else if(nbold[i]->isleftchildofvsplit(pi.imdim))
				lilold+= clil(sold[i].n,0,sold[i].sy,0.0,pi.sigma,pi.sigmaf,pi.tauc*pi.tauc,pi.muc);
			else
				lilold+= clil(sold[i].n,sold[i].nf,sold[i].sy,sold[i].sf,pi.sigma,pi.sigmaf,pi.tau0*pi.tau0,pi.mu0);
	}
	for(size_t i=0;i<nbnew.size();i++) {
      	if( (snew[i].n+snew[i].nf) >= minperbot ) {
			if(nbnew[i]->isrightchildofvsplit(pi.imdim)) //use prior on mu for bias case.  Here, sl.nf and sr.nf must be >=minperbot
				lilnew += clil(0,snew[i].nf,0.0,snew[i].sf,pi.sigma,pi.sigmaf,pi.taud*pi.taud,pi.mud);
			else if(nbnew[i]->isleftchildofvsplit(pi.imdim))  //use usual prior on mu
				lilnew += clil(snew[i].n,0,snew[i].sy,0.0,pi.sigma,pi.sigmaf,pi.tauc*pi.tauc,pi.muc);
			else
				lilnew += clil(snew[i].n,snew[i].nf,snew[i].sy,snew[i].sf,pi.sigma,pi.sigmaf,pi.tau0*pi.tau0,pi.mu0);
		}
		else hardreject=true;
	}

	double alpha1;
	alpha1=prinew*Qnew_to_old/priold/Qold_to_new/pm1/pm2*ps1*ps2;

    double alpha2 = alpha1*exp(lilnew-lilold);
	double alpha = std::min(1.0,alpha2);

	if(hardreject) alpha=0.0;

	if(gen.uniform()<alpha) {
		pi.taccept++;
		x = *tnew;
		return true;
	}
	else {
		return false;
	}

	return false;  // we never actually get here.
}


//--------------------------------------------------
// Rotate a given rotatable node in the tree.
// node n must be the left child of n->p and also not a root or terminal leaf node.
void rotright(tree::tree_p n)
{
	tree::tree_cp ctstar;
	tree::tree_p tstar;
	tree::tree_p newt = new tree;
	size_t vprime;
	size_t cprime;

	tstar=n->p->r;
	ctstar=n->p->r;
	tree::tree_p newnr = new tree(*ctstar);
	tstar->p=0;

	vprime=n->v;
	cprime=n->c;	
	n->v=n->p->v;
	n->c=n->p->c;
	n->p->v=vprime;
	n->p->c=cprime;
	newt->v=n->v;
	newt->c=n->c;

	newt->p=n->p;
	newt->l=n->r;
	newt->l->p=newt;
	newt->r=tstar;
	tstar->p=newt;
	n->p->r=newt;
	
	n->r=newnr;
	newnr->p=n;
}

void rotleft(tree::tree_p n)
{
	tree::tree_cp ctstar;
	tree::tree_p tstar;
	tree::tree_p newt = new tree;
	size_t vprime;
	size_t cprime;

	tstar=n->p->l;
	ctstar=n->p->l;
	tree::tree_p newnl = new tree(*ctstar);
	tstar->p=0;

	vprime=n->v;
	cprime=n->c;
	n->v=n->p->v;
	n->c=n->p->c;
	n->p->v=vprime;
	n->p->c=cprime;
	newt->v=n->v;
	newt->c=n->c;

	newt->p=n->p;
	newt->r=n->l;
	newt->r->p=newt;
	newt->l=tstar;
	tstar->p=newt;
	n->p->l=newt;

	n->l=newnl;
	newnl->p=n;
}

// reduce the left sub-tree of the node that was rotated to the top
void reduceleft(tree::tree_p n, size_t v, size_t c)
{
	tree::tree_p temp;

	if(n->r->l && n->r->v==v) //right is not terminal and splits on v
		if(n->r->c >= c) { //then only keep left branch
			delete n->r->r;
			temp=n->r;
			n->r=temp->l;
			temp->l->p=n;
			temp->r=0;
			temp->l=0;
			temp->p=0;
			delete temp;
		}
	if(n->l->l && n->l->v==v) //left is not terminal and splits on v
		if(n->l->c >= c) { // then only keep left branch
			delete n->l->r;
			temp=n->l;
			n->l=temp->l;
			temp->l->p=n;
			temp->r=0;
			temp->l=0;
			temp->p=0;
			delete temp;
		}
}

void reduceright(tree::tree_p n, size_t v, size_t c)
{
	tree::tree_p temp;

	if(n->r->v==v && n->r->l) //right is not terminal and splits on v
		if(n->r->c <= c) { //then only keep right branch
			delete n->r->l;
			temp=n->r;
			n->r=temp->r;
			temp->r->p=n;
			temp->r=0;
			temp->l=0;
			temp->p=0;
			delete temp;
		}
	if(n->l->v==v && n->l->l) //left is not terminal and splits on v
		if(n->l->c <= c) { // then only keep right branch
			delete n->l->l;
			temp=n->l;
			n->l=temp->r;
			temp->r->p=n;
			temp->r=0;
			temp->l=0;
			temp->p=0;
			delete temp;
		}
}

void splitleft(tree::tree_p t, size_t v, size_t c)
{
	tree::tree_p temp;

	if(t->l) //not terminal node
	{
		if(t->v==v && t->c >= c)
		{
			temp=t->l;
			if(t->isleft())
			{
				t->p->l=temp;
				temp->p=t->p;
			}
			else //isright
			{
				t->p->r=temp;
				temp->p=t->p;
			}
			delete t->r;
			t->p=0;
			t->r=0;
			t->l=0;
			delete t;
			t=temp;
			splitleft(t,v,c);
		}
		else
		{
			splitleft(t->l,v,c);
			splitleft(t->r,v,c);
		}
	}
}

void splitright(tree::tree_p t, size_t v, size_t c)
{
	tree::tree_p temp;

	if(t->l) //not terminal node
	{
		if(t->v==v && t->c <= c)
		{
			temp=t->r;
			if(t->isleft())
			{
				t->p->l=temp;
				temp->p=t->p;
			}
			else //isright
			{
				t->p->r=temp;
				temp->p=t->p;
			}
			delete t->l;
			t->p=0;
			t->l=0;
			t->r=0;
			delete t;
			t=temp;
			splitright(t,v,c);
		}
		else
		{
			splitright(t->l,v,c);
			splitright(t->r,v,c);
		}
	}
}


bool merge2(tree::tree_p tl, tree::tree_p tr, tree::tree_p t, size_t v, size_t c, RNG& gen)
{
	bool m1,m2;
	tree::tree_cp temptl,temptr;
	int tnwl=0,tnwr=0;
	double u;

	u=gen.uniform();

	if(arenodesleafs(tl,tr)) {  //merging type 3
		if(u<0.5) {
			t->v=tl->v;
			t->c=tl->c;
			t->mu=tl->mu;  //doesn't matter actually it will be overwritten.
			t->l=0;
			t->r=0;
		}
		else 
		{
			temptl=tl;
			temptr=tr;
			t->v=v;
			t->c=c;
			t->l=new tree(*temptl);
			t->r=new tree(*temptr);
			t->l->p=t;
			t->r->p=t;
		}
		return true;
	}
	else if(arenodesequal(tl,tr) && !splitsonv(tl,tr,v)) {  //merging type 4
		m1=merge3(tl->l,tr->l,v,c,&tnwl);
		m2=merge3(tl->r,tr->r,v,c,&tnwr);
		if(u < (1.0/(tnwl+tnwr+1.0)) )
		{
			temptl=tl;
			temptr=tr;
			t->v=v;
			t->c=c;
			t->l=new tree(*temptl);
			t->r=new tree(*temptr);
			t->l->p=t;
			t->r->p=t;
		}
		else
		{
			t->v=tl->v;
			t->c=tl->c;
			t->l=new tree;
			t->r=new tree;
			t->l->p=t;
			t->r->p=t;
			tnwl=0;
			tnwr=0;
			m1=merge2(tl->l,tr->l,t->l,v,c,gen);
			m2=merge2(tl->r,tr->r,t->r,v,c,gen);
		}
		return (m1 & m2);
	}
	else if(splitsonv(tl,tr,v)) {  //merging type 7
		m1=merge3(tl->r,tr,v,c,&tnwr);
		m2=merge3(tl,tr->l,v,c,&tnwl);
		if(u < (1.0/(tnwr+tnwl+1.0)) )
		{
			temptl=tl;
			temptr=tr;
			t->v=v;
			t->c=c;
			t->l=new tree(*temptl);
			t->r=new tree(*temptr);
			t->l->p=t;
			t->r->p=t;
		}
		else if(u < ((1.0+tnwr)/(1.0+tnwr+tnwl)) )
		{
			temptl=tl->l;
			t->v=tl->v;
			t->c=tl->c;
			t->l=new tree(*temptl);
			t->l->p=t;
			t->r=new tree;
			t->r->p=t;
			m2=merge2(tl->r,tr,t->r,v,c,gen);
		}
		else
		{
			temptr=tr->r;
			t->v=tr->v;
			t->c=tr->c;
			t->r=new tree(*temptr);
			t->r->p=t;
			t->l=new tree;
			t->l->p=t;
			m1=merge2(tl,tr->l,t->l,v,c,gen);
		}
		if(!m1) cout << "doh7a" << endl;
		if(!m2) cout << "doh7b" << endl;	
		return (m1 & m2);
	}
	else if(splitsonv(tl,v) && isleaf(tr)) //merging type 1
	{
		m1=merge3(tl->r,tr,v,c,&tnwr);
		if(u < (1.0/(tnwr + 1.0)) )
		{
			temptl=tl;
			temptr=tr;
			t->v=v;
			t->c=c;
			t->l=new tree(*temptl);
			t->r=new tree(*temptr);
			t->l->p=t;
			t->r->p=t;
		}
		else
		{
			temptl=tl->l;
			t->v=tl->v;
			t->c=tl->c;
			t->l=new tree(*temptl);
			t->l->p=t;
			t->r=new tree;
			t->r->p=t;
			m1=merge2(tl->r,tr,t->r,v,c,gen);
		}
		if(!m1) cout << "doh1(m1)" << endl;
		return m1;
	}
	else if(splitsonv(tr,v) && isleaf(tl)) //merging type 2
	{
		m2=merge3(tl,tr->l,v,c,&tnwl);
		if(u < (1.0/(tnwl+1.0)) )
		{
			temptl=tl;
			temptr=tr;
			t->v=v;
			t->c=c;
			t->l=new tree(*temptl);
			t->r=new tree(*temptr);
			t->l->p=t;
			t->r->p=t;			
		}
		else
		{
			temptr=tr->r;
			t->v=tr->v;
			t->c=tr->c;
			t->r=new tree(*temptr);
			t->r->p=t;
			t->l=new tree;
			t->l->p=t;
			m2=merge2(tl,tr->l,t->l,v,c,gen);
		}
		if(!m2) cout << "doh2(m2)" << endl;
		return m2;
	}
	else if(!isleaf(tl) && !isleaf(tr) && splitsonv(tr,v)) { //merge type 6(i)
		m1=merge3(tl,tr->l,v,c,&tnwr);
		if(u < (1.0/(tnwr + 1.0)) )
		{
			temptl=tl;
			temptr=tr;
			t->v=v;
			t->c=c;
			t->l=new tree(*temptl);
			t->r=new tree(*temptr);
			t->l->p=t;
			t->r->p=t;			
		}
		else
		{
			temptr=tr->r;
			t->v=tr->v;
			t->c=tr->c;
			t->r=new tree(*temptr);
			t->r->p=t;
			t->l=new tree;
			t->l->p=t;
			m1=merge2(tl,tr->l,t->l,v,c,gen);
		}
		if(!m1) cout << "doh6i(m1)" << endl;
		return m1;
	}
	else if(!isleaf(tl) && !isleaf(tr) && splitsonv(tl,v)) { //merge type 6(ii)
		m2=merge3(tl->r,tr,v,c,&tnwl);
		if(u < (1.0/(tnwl + 1.0)) )
		{
			temptl=tl;
			temptr=tr;
			t->v=v;
			t->c=c;
			t->l=new tree(*temptl);
			t->r=new tree(*temptr);
			t->l->p=t;
			t->r->p=t;
		}
		else
		{
			temptl=tl->l;
			t->v=tl->v;
			t->c=tl->c;
			t->l=new tree(*temptl);
			t->l->p=t;
			t->r=new tree;
			t->r->p=t;
			m2=merge2(tl->r,tr,t->r,v,c,gen);
		}
		if(!m2) cout << "doh6ii(m2)" << endl;
		return m2;
	}
	else if(!splitsonv(tl,v) && isleaf(tr)) { //merge type 5(i)
		temptl=tl;
		temptr=tr;
		t->v=v;
		t->c=c;
		t->l=new tree(*temptl);
		t->r=new tree(*temptr);
		t->l->p=t;
		t->r->p=t;
		return true;
	}
	else if(!splitsonv(tr,v) && isleaf(tl)) { //merge type 5(ii)
		temptl=tl;
		temptr=tr;
		t->v=v;
		t->c=c;
		t->l=new tree(*temptl);
		t->r=new tree(*temptr);
		t->l->p=t;
		t->r->p=t;
		return true;
	}
	else // default type aka type 8
	{
		temptl=tl;
		temptr=tr;
		t->v=v;
		t->c=c;
		t->l=new tree(*temptl);
		t->r=new tree(*temptr);
		t->l->p=t;
		t->r->p=t;
		return true;
	}

	return false;
}

// only to get nways, not to actually do the merge.
bool merge3(tree::tree_p tl, tree::tree_p tr, size_t v, size_t c, int* nways)
{
	bool m1,m2;
	int tnwl=0,tnwr=0;

	if(arenodesleafs(tl,tr)) {  //merging type 3
		*nways += 2;
		return true;
	}
	else if(arenodesequal(tl,tr) && !splitsonv(tl,tr,v)) {  //merging type 4
		*nways += 1;
		m1=merge3(tl->l,tr->l,v,c,&tnwl);
		m2=merge3(tl->r,tr->r,v,c,&tnwr);
		*nways += (tnwr*tnwl);
		return (m1 & m2);
	}
	else if(splitsonv(tl,tr,v)) {  //merging type 7
		*nways += 1;  //for this one
		m1=merge3(tl->r,tr,v,c,&tnwr);
		m2=merge3(tl,tr->l,v,c,&tnwl);
		*nways+= (tnwr+tnwl);
		if(!m1) cout << "doh7a" << endl;
		if(!m2) cout << "doh7b" << endl;	
		return (m1 & m2);
	}
	else if(splitsonv(tl,v) && isleaf(tr)) //merging type 1
	{
		*nways += 1; //for this one
		m1=merge3(tl->r,tr,v,c,&tnwr);
		*nways += tnwr;
		if(!m1) cout << "doh1(m1)" << endl;
		return m1;
	}
	else if(splitsonv(tr,v) && isleaf(tl)) //merging type 2
	{
		*nways += 1; //for this one
		m2=merge3(tl,tr->l,v,c,&tnwl);
		*nways += tnwl;
		if(!m2) cout << "doh2(m2)" << endl;
		return m2;
	}
	else if(!isleaf(tl) && !isleaf(tr) && splitsonv(tr,v)) { //merge type 6(i)
		*nways += 1; //for this one
		m1=merge3(tl,tr->l,v,c,&tnwr);
		*nways += tnwr;
		if(!m1) cout << "doh6i(m1)" << endl;
		return m1;
	}
	else if(!isleaf(tl) && !isleaf(tr) && splitsonv(tl,v)) { //merge type 6(ii)
		*nways +=1 ; //for this one
		m2=merge3(tl->r,tr,v,c,&tnwl);
		*nways += tnwl;
		if(!m2) cout << "doh6ii(m2)" << endl;
		return m2;
	}
	else if(!splitsonv(tl,v) && isleaf(tr)) { //merge type 5(i)
		*nways += 1; //for this one
		return true;
	}
	else if(!splitsonv(tr,v) && isleaf(tl)) { //merge type 5(ii)
		*nways += 1; //for this one
		return true;
	}
	else // default type aka type 8
	{
		*nways += 1; //for this one
		return true;
	}

	return false;
}

//--------------------------------------------------
// These ones support the rotate code, but could be generally useful too.
bool hasvcsplit(tree::tree_p t, size_t v, size_t c)
{
   tree::npv tnodes;

   tnodes.clear();
   t->getnodesonvc(tnodes,v,c);
   if(tnodes.size())
      return true;
   return false;
}

bool splitsonv(tree::tree_p t, size_t v)
{
   if(!t->getl()) //its a leaf
      return false;
   if(t->getv()==v) return true;
   return false;
}

bool splitsonv(tree::tree_p nl, tree::tree_p nr, size_t v)
{
   if(!nl->getl() || !nr->getl()) //one is a leaf
      return false;
   if(nl->getv()==v && nr->getv()==v) return true;
   return false;
}

bool isleaf(tree::tree_p t)
{
   if(!t->getl()) return true;
   return false;
}

bool arenodesequal(tree::tree_p nl, tree::tree_p nr)
{
   if(!nl->getl() || !nr->getl())  //one is a leaf
      return false;
   if(nl->getv()==nr->getv() && nl->getc()==nr->getc()) //both split on v at c
      return true;
   return false;
}

bool arenodesleafs(tree::tree_p nl, tree::tree_p nr)
{
   if(!nl->getl() && !nr->getl()) //both are leaves
      return true;
   return false;
}


